/*
  Glua: Lua-C interface glue - v 1.0

  Copyright (c) 2002 Enrico Colombini. All rights reserved.

  This software, including its documentation, is provided 'as-is',
  without any express or implied warranty. In no event will the author
  be held liable for any damages arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

  2. Altered source versions must be plainly marked as such, and must not
     be misrepresented as being the original software.

  3. This notice may not be removed or altered from any source distribution.

*/


#include <string.h>
#include <stdio.h>
#include "glua.h"


#define keybase "glua_tag_"  /* base for building registry tag key */


/*--- Prototypes of local functions -------------------------------*/

/* glua_maketagkey: build a registry key for a tag */

static char *glua_maketagkey(int tag);


/* glua_error: signal an error of the chosen type */

static void glua_error (lua_State *lua, const char *fname,
                        int argn, int type, const char *tname);


/*--- Public functions ----------------------------------------------*/

void glua_fregister (lua_State *lua, glua_fitem flist[], int size)
{
  while (size-- > 0) {
    lua_register(lua, flist->name, flist->func);
    flist++;
  }
}


void glua_cregister (lua_State *lua, glua_citem clist[], int size)
{
  while (size-- > 0) {
    lua_pushnumber(lua, clist->val);
    lua_setglobal(lua, clist->name);
    clist++;
  }
}


double glua_numerror (lua_State *lua, const char *fname, int argn)
{
  glua_error(lua, fname, argn, LUA_TNUMBER, "");
  return 0.0;
}


char* glua_strerror (lua_State *lua, const char *fname, int argn)
{
  glua_error(lua, fname, argn, LUA_TSTRING, "");
  return NULL;
}


void* glua_usrerror (lua_State *lua, const char *fname, int tag, int argn)
{
  char *tagkey = glua_maketagkey(tag);
  const char *tagname;

  /* clean Lua-C stack to prevent worst-case overflow */
  lua_settop(lua, 0);
  /* read tagname from Lua registry */
  lua_getregistry(lua);
  lua_pushstring(lua, tagkey);
  lua_rawget(lua, -2);
  /* if tag name not in the registry, use numeric part of tag key instead */
  if (lua_isnil(lua, -1)) {
    tagname = tagkey + strlen(keybase);
  } else {
    tagname = lua_tostring(lua, -1);
  }
  glua_error(lua, fname, argn, LUA_TUSERDATA, tagname);
  return NULL;
}


extern void glua_noneerror (lua_State *lua, const char *fname, int argn)
{
  glua_error(lua, fname, argn, LUA_TNONE, "");
}


int glua_regnewtag(lua_State *lua, char *tagname)
{
  int tag = lua_newtag(lua);  /* ask Lua for a new tag number */

  /* if no stack space (should not happen) do not write into registry */
  if (lua_stackspace(lua) >= 3) {
    /* write key-tagname into Lua registry */
    lua_getregistry(lua);
    lua_pushstring(lua, glua_maketagkey(tag));
    lua_pushstring(lua, tagname);
    lua_rawset(lua, -3);
    /* leave the stack as it was */
    lua_settop(lua, -2);
  }
  return tag;
}


/*--- Local functions ---------------------------------------------*/

/* glua_maketagkey: build a registry key for a tag
 *
 * the key will be in the form "keybase_xxx..." where
 * <keybase> is defined at the start of this file and
 * "xxx..." is a n-digit hex string representing the tag value
 * (n depends from the word size of the CPU)
 * returns a pointer to a static key string (reused at each call)
 */
static char *glua_maketagkey(int tag)
{
  static char tagkey[sizeof(keybase) + 2 * sizeof(int) + 1];

  sprintf(tagkey , "%s%0*x", keybase, 2 * sizeof(int), tag);
  return tagkey;
}


/* glua_error: build a human-readable error message and call lua_error
 *
 * If userdata is concerned, show expected tag name assigned with
 * glua_newtag, if any, else show tag numeric value (tag name
 * comes from calling function anyway).
 */
static void glua_error (lua_State *lua, const char *fname,
                        int argn, int type, const char *tname)
{
  static char errbuf[256];  /*will be checked for overflow*/
  char *msg;

  switch (type){
  case LUA_TNUMBER:
    msg = "%s(): arg #%d must be a valid number%s";
    break;
  case LUA_TSTRING:
    msg = "%s(): arg #%d must be a valid string%s";
    break;
  case LUA_TUSERDATA:
    msg = "%s(): arg #%d must be an userdata with tag <%s>";
    break;
  case LUA_TNONE:
    msg = "%s(): unexpected arg #%d%s";
    break;
  default:
    msg = "problem in glua_error (%s(), arg #%d)";
  }
  /* check if enough space in errbuf (fname + msg + arg# + tagname + '\0')*/
  if ((strlen(fname) + strlen(msg) + sizeof(int) + strlen(tname) + 1)
  > sizeof (errbuf)) {
    lua_error(lua, "glua_error(): errbuf overflow");
  } else {
    sprintf(errbuf, msg, fname, argn, tname);
    lua_error(lua, errbuf);
  }
}



