/*
  Glua: Lua-C interface glue - v 1.0

  Copyright (c) 2002 Enrico Colombini. All rights reserved.

  This software, including its documentation, is provided 'as-is',
  without any express or implied warranty. In no event will the author
  be held liable for any damages arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

  2. Altered source versions must be plainly marked as such, and must not
     be misrepresented as being the original software.

  3. This notice may not be removed or altered from any source distribution.

*/


#ifndef glua_h
#define glua_h


#include "lua.h"


/*--- Types -------------------------------------------------------*/

/* info to register a C function */

typedef struct {
  char *name;
  int (*func)(lua_State *lua);
} glua_fitem;


/* info to register a C constant */

typedef struct {
  char *name;
  double val;
} glua_citem;


/*--- Functions for glua internal use -----------------------------*/

/* Error functions (no macros to save an argument) */

extern double glua_numerror (lua_State *lua, const char *fname, int argn);
extern char*  glua_strerror (lua_State *lua, const char *fname, int argn);
extern void*  glua_usrerror (lua_State *lua, const char *fname, int tag, int argn);
extern void   glua_noneerror (lua_State *lua, const char *fname, int argn);

/* Registration functions */

extern void glua_fregister (lua_State *lua, glua_fitem flist[], int size);
extern void glua_cregister (lua_State *lua, glua_citem clist[], int size);
extern int  glua_regnewtag(lua_State *lua, char *tagname) ;


/*--- Macros ------------------------------------------------------*/

/* glua_newtag: request a Lua tag and register its name
 *
 * Get a new tag number from Lua, register its user-assigned name
 * in the Lua registry with the key "glua_tag_xxx..." for use
 * in case of error by the macro glua_getUserdata
 *
 * For this macro to work, <lua> must point
 * to the current Lua state
 */

#define glua_newtag(tagname)\
  glua_regnewtag(lua, tagname);


/* glua_regFunctions: register a list of functions
 *
 * The list must be an array of glua_fn items
 *
 * For this macro to work, <lua> must point
 * to the current Lua state
 */

#define glua_regFunctions(flist)\
  glua_fregister(lua, flist, sizeof(flist) / sizeof(glua_fitem))


/* glua_regConstants: register a list of numeric constants
 *
 * The list must be an array of glua_const items
 *
 * For this macro to work, <lua> must point
 * to the current Lua state
 */

#define glua_regConstants(clist)\
  glua_cregister(lua, clist, sizeof(clist) / sizeof(glua_citem))


/* glua_fn: write a function declaration line in a glua_flist
 *
 * Example:
 *   glua_fn(Line) will expand to:
 *
 *   "Line", glua_fn_Line
 */
#define glua_fn(fname)\
  { #fname, glua##_fn_##fname }


/* glua_const: write a constant declaration line in a glua_clist
 *
 * Example:
 *   glua_const(ZOT) will expand to:
 *
 *   "ZOT", ZOT
 */
#define glua_const(cname)\
  { #cname, cname }


/* glua_function: define a Lua-callable C function
 *
 * The function will be glua_fn_fname() and
 * the following variables will be defined:
 * <lua> pointer to the Lua state
 * <glua_fname> the function name as seen from Lua
 *
 * Example:
 *   glua_function(Line) will expand to:
 *
 *   static int glua_fn_Line(lua_State* lua) {
 *     static char *glua_fname = "Line";
 *
 * Note: do not add the usual opening brace '{'
 * after the function declaration
 *
 * Note: the same name of the C function to call
 * can be used without conflict.
 */
#define glua_function(fname)\
  static int glua##_fn_##fname(lua_State* lua)\
  {\
    static const char *glua_fname = #fname;


/* Retrieve a value from the Lua stack
 *
 * All of the following macros return a value of the
 * given type or issue an error message if the value
 * is not of the correct type (and tag, if applicable)
 *
 * glua_getNumber returns a double
 * glua_getInteger returns an int (just syntactic sugar)
 * glua_getString returns a char*
 * glua_getUserdata returns a void*
 *
 * glua_getNone must find no value
 * (useful to check for no extra arguments)
 *
 * For these macros to work, <lua> must point
 * to the current Lua state (this is automatic
 * if glua_function has been called)
 */
#define glua_getNumber(argn)\
  (lua_isnumber(lua, argn)) ?\
    lua_tonumber(lua, argn)\
    : glua_numerror(lua, glua_fname, argn)

#define glua_getInteger(argn)\
  (int)glua_getNumber(argn)

#define glua_getString(argn)\
  (lua_isstring(lua, argn)) ?\
    lua_tostring(lua, argn)\
    : glua_strerror(lua, glua_fname, argn)

#define glua_getUserdata(argn, tag)\
  (lua_isuserdata(lua, argn) && lua_tag(lua, argn) == tag) ?\
    lua_touserdata(lua, argn)\
    : glua_usrerror(lua, glua_fname, tag, argn)

#define glua_getNone(argn)\
  if (lua_gettop(lua) >= argn) {\
    glua_noneerror(lua, glua_fname, argn);\
  }


/* glua_pushPointer: push a pointer into the Lua stack
 *
 * If the pointer is not NULL it is pushed as userdata
 * of the given tag, else <nil> is pushed.
 *
 * This avoids problems with invalid (NULL) pointers
 * being passed back from Lua to C (and, besides, takes
 * care of a Lua 4.0 bug: pushing an userdata value of 0
 * gives an unpredictable value).
 *
 * The returned value can also be used as a Lua boolean.
 *
 * For this macro to work, <lua> must point
 * to the current Lua state (this is automatic
 * if glua_function has been called)
 */

#define glua_pushPointer(p, tag)\
  if (p != NULL) { lua_pushusertag(lua, p, tag); }\
  else { lua_pushnil(lua); }


/* glua_pushBoolean: push a numeric value or <nil>
 *
 * If the value is not 0 it is pushed, else <nil> is pushed.
 *
 * This is of course not a regular "boolean",
 * but can be be treated as such by Lua (at least until 4.0).
 *
 * For this macro to work, <lua> must point
 * to the current Lua state (this is automatic
 * if glua_function has been called)
 */

#define glua_pushBoolean(v)\
  if ((v) != 0) { lua_pushnumber(lua, v); }\
  else { lua_pushnil(lua); }


/* Push a value into a Lua stack
 *
 * This is just sintactic sugar to avoid the use of
 * the <lua> variable (in most cases) and have the code
 * look uniform.
 *
 * glua_pushNumber pushes a numeric value (int or double)
 * glua_pushString pushes a C string
 * glua_pushNil pushes <nil>
 *
 * For these macros to work, <lua> must point
 * to the current Lua state (this is automatic
 * if glua_function has been called)
 */

#define glua_pushNumber(v)\
  lua_pushnumber(lua, v)

#define glua_pushString(s)\
  lua_pushstring(lua, s)

#define glua_pushNil()\
  lua_pushnil(lua)


#endif

