/*
  Test file for glua 1.0 (C side)
  Copyright (c) 2002 Enrico Colombini.
*/


#include "glua.h"
#include <stdio.h>
#include <stdlib.h>


/*----- A minimal function: compute a bitwise AND -----*/

/* Lua: c = Bitand(a, b)  [a,b,c: numbers] */

glua_function(Bitand)
  int a = glua_getInteger(1);  /* get arg #1 */
  int b = glua_getInteger(2);  /* get arg #2 */
  glua_getNone(3);             /* error if extra args */
  glua_pushNumber(a & b);      /* compute and return 1 numeric result */
  return 1;
}


/*----- A function that receives a constant and returns a string -----*/

/* define constants (see also clist[] at the bottom) */

enum { GOOD = 0, BAD, DUBIOUS, INVALID };

/* local data */

static char *answers[] = {           /* strings corresponding to constants */
  "It works perfectly!",
  "There's no hope to make it work. I give up.",
  "Just a couple of small problems to fix.",
  "What are you meaning?"
};

/* Lua: s = StatusDesc(c)  [c: integer, s: string] */

glua_function(StatusDesc)
  int c = glua_getInteger(1);        /* arg #1 the constant */
  glua_getNone(2);                   /* error if extra arg */
  if (c < GOOD || c > INVALID) {     /* range check */
    c = INVALID;
  }
  glua_pushString(answers[c]);       /* return one string result */
  return 1;
}


/*----- Calling a C function of the same name -----*/

static int Next(int n) { return n + 1; }  /* a miniature C function */

/* Lua: m = Next(n)  [n.m: integers] */

glua_function(Next)
  int v = glua_getInteger(1);        /* arg #1 */
  glua_getNone(2);                   /* error if extra arg */
  glua_pushNumber(Next(v));          /* return one numeric result */
  return 1;
}


/*----- Working with pointers to C data -----*/

typedef struct {    /* declare a struct type */
  double x;
  double y;
} Coord;

int coord_tag;  /* tag for Coord userdata, see init at the bottom */
int what_tag;   /* tag for What, an error-testing userdata */

/* allocate a new Coord, return a C pointer (or nil) */

/* Lua: c = NewCoord()  [c: Coord] */

glua_function(NewCoord)
  double x = glua_getNumber(1);           /* arg #1 */
  double y = glua_getNumber(2);           /* arg #2 */
  Coord *cp;
  glua_getNone(3);                        /* error if extra arg */
  cp = (Coord *)malloc(sizeof(Coord));    /* allocate a Coord struct*/
  if (cp) {                               /* initialize members */
    cp->x = x;
    cp->y = y;
  }
  glua_pushPointer(cp, coord_tag);        /* return one Coord* result */
  return 1;
}

/* get Coord values (protection against nil is automatic) */

/* Lua: x,y = GetCoord(c)  [c: Coord, x,y: numbers] */

glua_function(GetCoord)
  Coord *cp = glua_getUserdata(1, coord_tag); /* arg #1 is a Coord pointer */
  glua_getNone(2);                            /* error if extra arg */
  glua_pushNumber(cp->x);                     /* return 2 numeric results */
  glua_pushNumber(cp->y);
  return 2;
}

/* delete a Coord (no automatic deallocation here) */

/* Lua: DeleteCoord()  [c: Coord] */

glua_function(DeleteCoord)
  Coord *cp = glua_getUserdata(1, coord_tag); /* arg #1 is a Coord pointer */
  glua_getNone(2);                            /* error if extra arg */
  free(cp);
  return 0;
}

/* allocate a What, for error testing */

/* Lua: w = NewWhat()  [c: What] */

glua_function(NewWhat)
  void *w = (void*)0xdeadbeef;            /* an useless pointer */
  glua_getNone(1);                        /* error if any arg */
  glua_pushPointer(w, what_tag);          /* return one What result */
  return 1;
}


/*----- Function whose behavior depends on arg type -----*/

/* Lua: r = Arg2(a,b)  [a,b: numbers, r:string] */
/*      r = Arg2(s)    [s: string, r:string]    */

glua_function(Arg2)
  double a, b;
  const char *s;
  if (lua_isnumber(lua, 1)) {             /* if arg #1 is a number */
    a = glua_getNumber(1);                /* get two numbers */
    b = glua_getNumber(2);
    glua_getNone(3);                      /* error if third arg */
    /* ... do some work with a,b ...*/
    glua_pushString("Got 2 numbers");     /* return 1 string result */
  } else {
    s = glua_getString(1);                /* arg #1 must be a string */
    glua_getNone(2);                      /* error if second arg */
    /* ... do some work with s...*/
    glua_pushString("Got a string");      /* return 1 string result */  }
  return 1;
}


/*----- List of functions visible from Lua -----*/

static glua_fitem flist[] = {
  glua_fn(Bitand),
  glua_fn(StatusDesc),
  glua_fn(Next),
  glua_fn(NewCoord),
  glua_fn(GetCoord),
  glua_fn(DeleteCoord),
  glua_fn(NewWhat),
  glua_fn(Arg2)
};


/*----- List of constants visible from Lua -----*/

static glua_citem clist[] = {
  glua_const(GOOD),
  glua_const(BAD),
  glua_const(DUBIOUS),
  glua_const(INVALID)
};


/*----- Initialization and tag creation -----*/

void gluatest_open (lua_State* lua)
{
  glua_regFunctions(flist);               /* functions */
  glua_regConstants(clist);               /* constants */
  coord_tag = glua_newtag("Coord");       /* tags */
  what_tag = glua_newtag("What?");
}



